#define ROOT_UID                  0

#define _BSD_SOURCE                  /* for setgroups() and other calls */
#include <stdlib.h>
#include <signal.h>
#include <termios.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/file.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <errno.h>
#include <getopt.h>
#include "../include/su_indep.h"

/* -------------------------------------------- */
/* ------ declarations ------------------------ */
/* -------------------------------------------- */

int is_terminal = 0;

/* -------------------------------------------- */
/* ------ some local (static) functions ------- */
/* -------------------------------------------- */

/* ------ terminal mode ----------------------- */

struct termios stored_mode;        /* initial terminal mode settings */

/* should be called once at the beginning */
void store_terminal_modes()
{
    if (isatty(STDIN_FILENO)) {
	is_terminal = 1;
	if (tcgetattr(STDIN_FILENO, &stored_mode) != 0) {
	    (void) fprintf(stderr, "su: couldn't copy terminal mode");
	    exit(1);
	}
    } else if (getuid()) {
	(void) fprintf(stderr, "su: must be run from a terminal\n");
	exit(1);
    } else 
	is_terminal = 0;
}

/*
 * Returns:
 *   0     ok
 * !=0     error
 */
int reset_terminal_modes()
{
    if (is_terminal && tcsetattr(STDIN_FILENO, TCSAFLUSH, &stored_mode) != 0) {
	(void) fprintf(stderr, "su: cannot reset terminal mode: %s\n"
		       , strerror(errno));
	return 1;
    }else
	return 0;
}

/* ------ unexpected signals ------------------ */

struct sigaction old_int_act, old_quit_act, old_tstp_act, old_pipe_act;

void disable_terminal_signals()
{
    /* 
     * Protect the process from dangerous terminal signals.
     * The protection is implemented via sigaction() because
     * the signals are sent regardless of the process' uid.
     */
    struct sigaction act;

    act.sa_handler = SIG_IGN;  /* ignore the signal */
    sigemptyset(&act.sa_mask); /* no signal blocking on handler
				  call needed */
    act.sa_flags = SA_RESTART; /* do not reset after first signal
				  arriving, restart interrupted
				  system calls if possible */
    sigaction(SIGINT, &act, &old_int_act);
    sigaction(SIGQUIT, &act, &old_quit_act);
    /*
     * Ignore SIGTSTP signals. Why? attacker could otherwise stop
     * a process and a. kill it, or b. wait for the system to
     * shutdown - either way, nothing appears in syslogs.
     */
    sigaction(SIGTSTP, &act, &old_tstp_act);
    /*
     * Ignore SIGPIPE. The parent `su' process may print something
     * on stderr. Killing of the process would be undesired.
     */
    sigaction(SIGPIPE, &act, &old_pipe_act);
}

void enable_terminal_signals()
{
    sigaction(SIGINT, &old_int_act, NULL);
    sigaction(SIGQUIT, &old_quit_act, NULL);
    sigaction(SIGTSTP, &old_tstp_act, NULL);
    sigaction(SIGPIPE, &old_pipe_act, NULL);
}

/* ------ terminal ownership ------------------ */

uid_t terminal_uid = (uid_t) -1;

/*
 * Change the ownership of STDIN if needed.
 * Returns:
 *   0     ok,
 *  -1     fatal error (continue of the work is impossible),
 *   1     non-fatal error.
 * In the case of an error "err_descr" is set to the error message
 * and "callname" to the name of the failed call.
 */
int change_terminal_owner(uid_t uid, int is_login, 
	  const char **callname, const char **err_descr)
{
    /* determine who owns the terminal line */
    if (is_terminal && is_login) {
	struct stat stat_buf;

	if (fstat(STDIN_FILENO,&stat_buf) != 0) {
            *callname = "fstat to STDIN";
	    *err_descr = strerror(errno);
	    return -1;
	}
	if(fchown(STDIN_FILENO, uid, -1) != 0) {
	    *callname = "fchown to STDIN";
            *err_descr = strerror(errno);
	    return 1;
	}
	terminal_uid = stat_buf.st_uid;
    }
    return 0;
}

void restore_terminal_owner()
{
    if (terminal_uid != (uid_t) -1) {
        if(fchown(STDIN_FILENO, terminal_uid, -1) != 0) {
	    syslog(LOG_ALERT
		    , "Terminal owner hasn\'t been restored: %s"
		    , strerror(errno));
        }
        terminal_uid = (uid_t) -1;
    }
}

/* ------ Process Death Control (tm) ---------- */

uid_t invoked_uid;

/*
 * Make the process unkillable by the user invoked it.
 * Returns:
 *   0     ok,
 *  -1     fatal error (continue of the work is impossible),
 *   1     non-fatal error.
 * In the case of an error "err_descr" is set to the error message
 * and "callname" to the name of the failed call.
 */
int make_process_unkillable(const char **callname
        , const char **err_descr)
{
    invoked_uid = getuid();
    if(setuid(geteuid()) != 0) {
        *callname = "setuid";
	*err_descr = strerror(errno);
	return -1;
    }else
	return 0;
}

int make_process_killable()
{
    return setreuid(invoked_uid, -1);
}

/* ------ command line parser ----------------- */

__attribute__ ((__noreturn__))
static void usage (int status)
{
	fprintf (stderr, "usage: su [-|-l] [-c \"command\"] [-s \"shell\"] [username]\n");
	exit (status);
}

void parse_command_line(int argc, char *argv[], su_context *ctx)
{
	int optc;
	struct option const longopts[] =
	{
		{ "command", required_argument, 0, 'c' },
		{ "login", no_argument, 0, 'l' },
		{ "shell", required_argument, 0, 's' },
		{ "help", no_argument, 0, 'h' },
		{ 0, 0, 0, 0 }
	};

	if (argc < 1)
		usage (EXIT_FAILURE);

	memset (ctx, 0, sizeof(*ctx));

	while ((optc = getopt_long (argc, argv, "c:ls:", longopts, 0)) != -1)
	{
		switch (optc)
		{
			case 'c':
				if (ctx->command)
				    usage (EXIT_FAILURE);
				ctx->command = optarg;
				break;

			case 'l':
				if (ctx->login)
				    usage (EXIT_FAILURE);
				ctx->login = 1;
				break;

			case 's':
				if (ctx->shell)
				    usage (EXIT_FAILURE);
				ctx->shell = optarg;
				break;

			case 'h':
				usage (EXIT_SUCCESS);

			default:
				usage (EXIT_FAILURE);
		}
	}

	if (optind < argc && !strcmp (argv[optind], "-"))
	{
		if (ctx->login)
		    usage (EXIT_FAILURE);
		ctx->login = 1;
		++optind;
	}

	if (optind < argc)
		ctx->user = argv[optind++];

	if (optind < argc)
	    usage (EXIT_FAILURE);

	if (!ctx->user)
	{	/* default user is superuser */
		const struct passwd *pw = getpwuid (ROOT_UID);
		if (!(pw && pw->pw_name))	/* No ROOT_UID!? */
		{
			fprintf (stderr, "su: no access to superuser identity (%d)!\n", ROOT_UID);
			exit (EXIT_FAILURE);
		}

		ctx->user = strdup (pw->pw_name);
		endpwent();
	}

	if (ctx->shell)
	{
		if (getuid())
		{
			fprintf (stderr, "su: setting shell: permission denied\n");
			exit (EXIT_FAILURE);
		}
		if (access (ctx->shell, X_OK) < 0)
		{
			fprintf (stderr, "su: setting shell: %s\n", strerror (errno));
			exit (EXIT_FAILURE);
		}
	}
}
